pro crosstalkgong,param,x,Cout
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; NAME:
;	crosstalk 
; CALLING SEQUENCE:
;	crosstalklg,param,xdata,Cout
; PURPOSE:
;	Compute the covariance matrix of the m,nu diagram
; INPUTS:
;	param		input parameters in the following order
;			(frequency,linewidth,splitting_i(5),2*l+1 amplitudes,6 noises,1 assymetry (optional))
;	x 		frequency of the points centered on zero (Npoints)
; OPTIONAL KEYWORDS:
; 	none
; OUTPUTS:
;	Cout 		Covariance matrix of the mode plus noise
; OPTIONAL OUTPUT:
;	none
; EXAMPLE:
;
; LIMITATIONS:
;	Here the crosstalk is assumed to be real (and signed).
;	It is the same for both the real and imaginary.
;	It is NOT the same for the signal and the noise
; COMMONS:
;	These many commons are needed to minimize the number of commons in the likelihood functions
;	ctnm1		leakage matrix (2*l+1,2*l+1)
;	croiseen	noise covariance matrix (2*l+1,2*l+1,nb_bruit)
;	nb_bruit	number of pixel noise (typically 3 to 6)	
;	profilename	name of the line profile with 3 parameters (amplitude, frequency, linewidth) 
;		        + 1 optional parameters (assymmetry)
; PROCEDURES USED:
;       p1, p2, p3 ,p4, p5, function_rot
; MODIFICATION HISTORY:
;	Beta 1: WRITTEN, Thierry Appourchaux, July 10, 1998
;		Derived from crosstalklg
;------------------------------------------------------------------------------

	common speed,Neval
        common crosstalk,ctnm1
        common matbruit,croiseen
        common nb_bruit, nbbruit
	common profile,profilename



N=N_elements(ctnm1(*,0))
degree=(N-1)/2
Npoints=N_elements(x)

Cout=dblarr(2*degree+1,2*degree+1,Npoints)

Nparam=N_elements(param)




;------------------------ donne des noms plus cools pour les parametres


freq=param(0)
width=param(1)
a_i=param(2:6)			;a_i

ampli=param(7:2*degree+7)

bruit=exp(param(2*degree+8:2*degree+13))


if (profilename ne 'lorentzian') then begin
	assy=param(2*degree+14)		;assymmetry
	paramline=dblarr(9)
	paramline(7)=assy
	paramline(8)=0.
endif else begin
	paramline=dblarr(7)
endelse

;------------------- noise
Cout_simple=dblarr(2*degree+1,2*degree+1)

for p=0,nbbruit-1 do begin
        Cout_simple(*,*) = Cout_simple(*,*) + reform(croiseen(p,*,*)*bruit(p)^2)
endfor	

Cout_noise=rebin(Cout_simple,2*degree+1,2*degree+1,Npoints)

f=systime(1)

;-----------------compute variances for each mode
variance=dblarr(Npoints,2*degree+1)


; fake crosstalk here at 1.0
paramline(5)=0.d0

paramline(2)=width

maxp= 2*degree < 5

;aa=systime(1)

for i=0,2*degree do begin
        m=i-degree
        paramline(0)=ampli(i)
        paramline(1)=freq
	for icg=1,maxp do begin

	zz='p'+strtrim(icg,1)

	s=call_function(zz,m*1.,degree*1.)

		paramline(1)=paramline(1)+a_i(icg-1)*degree*s
	endfor 
	
	variance(*,i)=call_function(profilename,paramline,x(*))


endfor


; In case you have a B angle...

	mat_rot=function_rot(degree,0.)

	rctnm=transpose(ctnm1 ## mat_rot)

; If B=0. then rctnm=transpose(ctnm1)
	
rrCout=dblarr(N,N,Npoints)

zz=dblarr(N,N)


rctnmk=dblarr(N,N)

varki=dblarr(N,N,Npoints)

for k=0,N-1 do begin
	zz(*,*)=0.
	zz(k,k)=1.d0
	rctnmk=transpose(rctnm) ## (zz ## rctnm)
	rctnmki=rebin(rctnmk,N,N,Npoints)

	varki(0,0,*)=reform(variance(0:Npoints-1,k))

	varki=rebin(varki(0,0,*),N,N,Npoints)
	
	rrCout=rrCout+varki*rctnmki
	
endfor	




rrCout=rrCout+Cout_noise



Cout(0:2*degree,0:2*degree,*)=rrCout(0:2*degree,0:2*degree,*)




Neval=Neval+1


end




;************************************************
; Compute (lorentzian) profile of one line
;************************************************

Function lorentzian,param,x

; This routine computes for a given  array x 
;
; the y profile :  k_A* A / ( 1 + (B-x+k_D*D)^2/C^2 )  + E
;
; remark : k_A and k_D are not variables in this routine
; k_D is for the azimutal order m

A   = exp(param(0)) 	; the amplitude
B   = param(1) 		; the central frequency of the m mode
C   = exp(param(2))/2. 	; the half linewidth
;D   = param(3) 		; the splitting
;E   = exp(param(4)) 	; the noise
k_A = exp(param(5)) 		; the crosstalk
alias=param(6)



; the value is computed in x
	z = 1. / ( 1.0d0 + ((B-x)/C)^2 )

	y = k_A * A * z 


;	if (alias ne 0.) then begin
;		y = y+alias*k_A * A / ( 1.0d0 + ((B-x-11.57)/C)^2 )
;		y = y+alias*k_A * A / ( 1.0d0 + ((B-x+11.57)/C)^2 )
;	endif
	

	return,y

end

;************************************************


Function lorentzian_ass,param,x

; This routine computes for a given  array x 
;
; the y ASSYMETRICAL profile derived from lorentzian
;
; remark : k_A is not variable in this routine


A   = exp(param(0)) 	; the amplitude
B   = param(1) 		; the central frequency of the m mode
C   = exp(param(2))/2. 	; the half linewidth
;D   = param(3) 		; the splitting
;E   = exp(param(4)) 	; the noise
k_A = exp(param(5)) 		; the crosstalk
alias=param(6)



; the value is computed in x
	z = 1. / ( 1.0d0 + ((B-x)/C)^2 )

	y = z 

	Basy = param(7)			; Amplitude of assymetry
;	wassy = param(8)

	zasy=1.-((Basy+B-x)/C)^2
		
	y = y * zasy

	y = k_A * A * (y - min(y))/(max(y)-min(y))	

	return, y 

end

;************************************************


Function lorentzian_rak,param,x

; This routine computes for a given  array x 
;
; the y ASSYMETRICAL profile derived from lorentzian
;
; remark : k_A is not variable in this routine


A   = exp(param(0)) 	; the amplitude
B   = param(1) 		; the central frequency of the m mode
C   = exp(param(2))/2. 	; the half linewidth
;D   = param(3) 		; the splitting
;E   = exp(param(4)) 	; the noise
k_A = exp(param(5)) 		; the crosstalk
alias=param(6)



; the value is computed in x
	z = 1. / ( 1.0d0 + ((B-x)/C)^2 )

	y = z 

	Basy = param(7)			; Amplitude of assymetry
;	wassy = param(8)

	zasy=(1.+(Basy*(x-B)/C))^2+Basy^2
		
	y = y * zasy

	y = k_A * A * y

;(y - min(y))/(max(y)-min(y))	

	return, y 

end


;************************************************


Function function_rot,l,beta
        dim=2*l+1
        mat=dblarr(dim,dim)

	for i=0,l do begin
 	    for j=-i,i do begin
 	          mat(i+l,j+l)=dmm(l,i,j,beta)
 	    endfor
	endfor

	for i=-l,0 do begin
	    for j=i,-i do begin
 	          mat(i+l,j+l)=mat(-i+l,-j+l)*(-1.d0)^(i-j)
  	    endfor
	endfor

	for j=0,l do begin
	    for i=-j,j do begin
 	          mat(i+l,j+l)=dmm(l,j,i,-beta)
	    endfor
	endfor

	for j=-l,0 do begin
 	    for i=j,-j do begin
	          mat(i+l,j+l)=mat(-i+l,-j+l)*(-1.d0)^(i-j)
	    endfor
	endfor

	return,mat
end


;************************
; compute dm1,m2(beta)  for  m1+m2>=0  and m1-m2>=0
;**************************
Function dmm,l,m1,m2,beta

        co=cos(beta/2.d0)
        si=sin(beta/2.d0)

        sum=0.d0

        for s=0,l-m1 do begin
             var=0.d0
             var=combi(l+m2,l-m1-s)*combi(l-m2,s)*(-1.d0)^(l-m1-s)
             var=var*co^(2*s+m1+m2)*si^(2*l-2*s-m1-m2)
             sum=sum+var
        endfor

        sum=sum*sqrt(factorial(l+m1)*factorial(l-m1)*1.d0)
        sum=sum/sqrt(factorial(l+m2)*factorial(l-m2)*1.d0)

        return,sum
end

;************************************************
; Compute combinations n r  ,    n>r
;************************************************
Function combi,n,r
        c=factorial(n)/factorial(n-r)/factorial(r)
        return,c
end

;************************************************
; Clebsh-Gordan coefficients
;************************************************


function p1,m,l
	return,(1.d0*m)/(1.d0*l)
end

function p2,m,l
	L2=l*(l+1)
	norm=(6.d0*l^2-2.d0*L2)
	return,(6.d0*m^2-2.d0*L2)/norm
end

function p3,m,l
	L2=l*(l+1)
	norm=(20.d0*l^3-4.d0*(3.d0*L2-1.d0)*l)
	return,(20.d0*m^3-4.d0*(3.d0*L2-1.d0)*m)/norm
end

function p4,m,l
	L2=l*(l+1)
	norm=(70.d0*l^4-10.d0*(6.d0*L2-5)*l^2+6.d0*L2*(L2-2.d0))
	return,(70.d0*m^4-10.d0*(6.d0*L2-5)*m^2+6.d0*L2*(L2-2.d0))/norm
end

function p5,m,l
	L2=l*(l+1)
	norm=(252.d0*l^5-140.d0*(2.d0*L2-3.d0)*l^3+(20.d0*L2*(3.d0*L2-10.d0)+48.d0)*l)
	return,(252.d0*m^5-140.d0*(2.d0*L2-3.d0)*m^3+(20.d0*L2*(3.d0*L2-10.d0)+48.d0)*m)/norm
end
